/* $USAGI: mipv6_ha_ipsec.c,v 1.2 2003/08/02 02:54:12 takamiya Exp $ */
/*
 * Copyright (C)2001 USAGI/WIDE Project
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *
 * Authors:
 *   Noriaki Takamiya <takamiya@linux-ipv6.org> / USAGI
 */
/*
 * from draft-ietf-mobileip-mipv6-ipsec-ha-06.txt:
 *
 * for Home Registration
 * 
 * o SPD1: SRC=HoA DST=HA  transport mode proto = MH
 * o SPD2: SRC=HA  DST=HoA transport mode proto = MH
 * o SAD1: SRC=HoA DST=HA  proto = MH
 * o SAD2: SRC=HA  DST=HoA proto = MH
 * 
 * for Return Routability
 * 
 * o SPD3: SRC=HoA DST=any tunnel mode SGW=HA  proto = MH
 * o SPD4: SRC=any DST=HoA tunnel mode SGW=CoA proto = MH
 * o SAD3: SRC=CoA DST=HA
 * o SAD4: SRC=HA  DST=CoA
 * 
 * for Mobile Prefix Discovery
 * 
 * o SPD5: SRC=HoA DST=HA  transport mode proto = ICMPv6
 * o SPD6: SRC=HA  DST=HoA transport mode proto = ICMPv6
 * o SAD5: SRC=HoA DST=HA  proto = ICMPv6
 * o SAD6: SRC=HA  DST=HoA proto = ICMPv6
 * 
 * for Payload Packet
 * 
 * o SPD7: SRC=HoA DST=any tunnel mode SGW=HA  proto = any
 * o SPD8: SRC=any DST=HoA tunnel mode SGW=CoA proto = any
 * o SAD7: SRC=CoA DST=HA
 * o SAD8: SRC=HA  DST=CoA
 *
 * This code changes SAD3, SAD4, SPD4, SAD7, SAD8, SPD8.
 */

#include <linux/config.h>
#include <linux/sched.h>
#include <linux/notifier.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/socket.h>
#include <linux/inet.h>
#include <linux/in6.h>
#include <linux/pfkeyv2.h>
#include <net/spd.h>
#include <net/sadb.h>
#include <linux/list.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <asm/softirq.h>
#include <linux/spinlock.h>
#include <asm/uaccess.h>
#include <net/mipv6.h>

#include "bul.h"
#include "bcache.h"
#include "mipv6_ha_ipsec.h"

/*
 * mipv6_change_sa_index: change the address of sa_index;
 */
int mipv6_change_sa_index(struct sa_index *sa, struct in6_addr *coa, __u8 dir)
{
	__u8 buf[INET6_ADDRSTRLEN];
	struct sockaddr_in6 *addr;
	struct ipsec_sa *sa_entry;

	if (!sa) {
		return 0;
	}

	/*
	 * Then, correspondent SADB will be changed
	 */
#ifdef CONFIG_IPV6_MOBILITY_DEBUG
		printk(KERN_INFO "%s(%d): spi = %x\n", __FUNCTION__, __LINE__, __constant_ntohl(sa->spi));
#endif
	sa_entry = sa->sa;
	if (!sa_entry) {
#ifdef CONFIG_IPV6_MOBILITY_DEBUG
		printk(KERN_INFO "%s: sa_index->sa is NULL\n", __FUNCTION__);
#endif
                sa_entry = sadb_find_by_sa_index(sa);
	}
	if (sa_entry) {
#ifdef CONFIG_IPV6_MOBILITY_DEBUG
		printk(KERN_INFO "%s: ipsec_sa is found.\n", __FUNCTION__);
#endif
		/*
		 * modify the dst address of ipsec_sa
		 */
		if (dir == MIPV6_IPSEC_HA_OUT) {
			addr = (struct sockaddr_in6 *)&sa_entry->dst;
#ifdef CONFIG_IPV6_MOBILITY_DEBUG
			printk(KERN_INFO "%s(%d): spi = %x\n", __FUNCTION__, __LINE__, __constant_ntohl(sa_entry->spi));
			printk(KERN_INFO "%s(%d): addr = %s\n", __FUNCTION__, __LINE__, in6_ntop(&addr->sin6_addr, buf));
			printk(KERN_INFO "%s(%d): coa = %s\n", __FUNCTION__, __LINE__, in6_ntop(coa, buf));
#endif
			memcpy(&addr->sin6_addr, coa, sizeof(struct in6_addr));
		} else if (dir == MIPV6_IPSEC_HA_IN) {
			addr = (struct sockaddr_in6 *)&sa_entry->src;
#ifdef CONFIG_IPV6_MOBILITY_DEBUG
			printk(KERN_INFO "%s(%d): spi = %x\n", __FUNCTION__, __LINE__, __constant_ntohl(sa_entry->spi));
			printk(KERN_INFO "%s(%d): addr = %s\n", __FUNCTION__, __LINE__, in6_ntop(&addr->sin6_addr, buf));
			printk(KERN_INFO "%s(%d): coa = %s\n", __FUNCTION__, __LINE__, in6_ntop(coa, buf));
#endif
			memcpy(&addr->sin6_addr, coa, sizeof(struct in6_addr));
		}
	} else {
#ifdef CONFIG_IPV6_MOBILITY_DEBUG
		printk(KERN_INFO "%s: ipsec_sa is NULL\n", __FUNCTION__);
#endif
        }

	/*
	 * modify the dst address of sa_index
	 */
	if (dir == MIPV6_IPSEC_HA_OUT) {
		addr = (struct sockaddr_in6 *)&sa->dst;
		memcpy(&addr->sin6_addr, coa, sizeof(struct in6_addr));
	}
	return 1;
}

/*
 * make selector for IPsec SPD
 * This function returns tunnel mode IPsec SPD specified by the Home Address,
 * Care-of Address and protocol.
 */
struct ipsec_sp  *mipv6_find_spd(struct in6_addr *dst, struct in6_addr *src,
	__u8 prefixlen_d, __u8 prefixlen_s, __u8 proto, __u8 mode)
{
	struct ipsec_sp *policy = NULL;
	struct selector sel;
	struct sockaddr_in6 *selector_dst;
	struct sockaddr_in6 *selector_src;
	__u8 buf[INET6_ADDRSTRLEN];

	if (!dst) {
#ifdef CONFIG_IPV6_MOBILITY_DEBUG
		printk(KERN_DEBUG "%s(%d): Destination Addressis NULL!\n", __FUNCTION__, __LINE__);
#endif
		return policy;
	}
	if (!src) {
#ifdef CONFIG_IPV6_MOBILITY_DEBUG
		printk(KERN_DEBUG "%s(%d): Source Address is NULL!\n", __FUNCTION__, __LINE__);
#endif
		return policy;
	}
#ifdef CONFIG_IPV6_MOBILITY_DEBUG
	printk(KERN_INFO "%s(%d): dst = %s\n", __FUNCTION__, __LINE__, in6_ntop(dst, buf));
	printk(KERN_INFO "%s(%d): src = %s\n", __FUNCTION__, __LINE__, in6_ntop(src, buf));
#endif
	memset(&sel, 0, sizeof(struct selector));

	/* destination address of target SPD is Home Address */
	selector_dst = (struct sockaddr_in6 *)&(sel.dst);
	sel.prefixlen_d = prefixlen_d;
	selector_dst->sin6_family = AF_INET6;
	memcpy(&selector_dst->sin6_addr, dst, sizeof(struct in6_addr));

	/* source address of target SPD is unspecified address */
	selector_src = (struct sockaddr_in6 *)&(sel.src);
	sel.prefixlen_s = prefixlen_s;
	memcpy(&selector_src->sin6_addr, src, sizeof(struct in6_addr));
	selector_src->sin6_family = AF_INET6;
	
	/* other fields */
	sel.proto = proto;
	sel.mode = mode;

	/* find the selector */
#if 0
	return ipsec_sp_get(&sel);
#else
	spd_find_by_selector(&sel, &policy);
	return policy;
#endif
}

/*
 * change SADB/SPD entry for Reverse Tunneling
 * This function 
 */
int __mipv6_mn_change_tunnel_ipsec_by_proto(struct mipv6_bul_entry *entry, __u8 proto)
{
	struct ipsec_sp *policy = NULL;

	struct in6_addr dst;
	struct in6_addr src;
	struct in6_addr *coa = &entry->coa;

	/*
	 * Phase 1: Change the following SA/SPD
	 *
	 * o SPD4: SRC=any DST=HoA tunnel mode SGW=CoA proto = MH
	 * o SPD8: SRC=any DST=HoA tunnel mode SGW=CoA proto = any
	 * o SAD4: SRC=HA  DST=CoA
	 * o SAD8: SRC=HA  DST=CoA
	 */ 
	memcpy(&dst, &entry->home_addr, sizeof(struct in6_addr));
	memset(&src, 0, sizeof(struct in6_addr));

#ifdef CONFIG_IPV6_MOBILITY_DEBUG
{
	__u8 buf[INET6_ADDRSTRLEN];
	printk(KERN_INFO "%s(%d): dst = %s\n", __FUNCTION__, __LINE__, in6_ntop(&dst, buf));
	printk(KERN_INFO "%s(%d): src = %s\n", __FUNCTION__, __LINE__, in6_ntop(&src, buf));
	printk(KERN_INFO "%s(%d): proto = %d\n", __FUNCTION__, __LINE__, proto);
}
#endif

	policy = mipv6_find_spd(&dst, &src, 128, 0, proto, IPSEC_MODE_TUNNEL);
	if (policy) {
		/*
		 * change the policy:
		 * 1) change the SGW to the current coa;
		 * 2) change the SADB destination address;
		 */
		write_lock_bh(&policy->lock);

		mipv6_change_sa_index(policy->auth_sa_idx, coa, MIPV6_IPSEC_HA_OUT);
		mipv6_change_sa_index(policy->esp_sa_idx, coa, MIPV6_IPSEC_HA_OUT);
		mipv6_change_sa_index(policy->comp_sa_idx, coa, MIPV6_IPSEC_HA_OUT);

		/* clean up */
		write_unlock_bh(&policy->lock);
		ipsec_sp_put(policy);
	} else {
#ifdef CONFIG_IPV6_MOBILITY_DEBUG
		printk(KERN_INFO "%s(%d): policy is NULL.\n", __FUNCTION__, __LINE__);
#endif
	}
	/*
	 * Phase 2: Change the following SA/SPD
	 *
	 * o SPD3: SRC=HoA DST=any tunnel mode SGW=HA  proto = MH
	 * o SPD7: SRC=HoA DST=any tunnel mode SGW=HA  proto = any
	 * o SAD3: SRC=CoA DST=HA
	 * o SAD7: SRC=CoA DST=HA
	 */
	policy = NULL;
	memset(&dst, 0, sizeof(struct in6_addr));
	memcpy(&src, &entry->home_addr, sizeof(struct in6_addr));

#ifdef CONFIG_IPV6_MOBILITY_DEBUG
{
	__u8 buf[INET6_ADDRSTRLEN];
	printk(KERN_INFO "%s(%d): dst = %s\n", __FUNCTION__, __LINE__, in6_ntop(&dst, buf));
	printk(KERN_INFO "%s(%d): src = %s\n", __FUNCTION__, __LINE__, in6_ntop(&src, buf));
	printk(KERN_INFO "%s(%d): proto = %d\n", __FUNCTION__, __LINE__, proto);
}
#endif
	policy = mipv6_find_spd(&dst, &src, 0, 128, proto, IPSEC_MODE_TUNNEL);
	if (policy) {
		/*
		 * change the policy:
		 * 1) change the SGW from the current coa;
		 * 2) change the SADB destination address;
		 */

		write_lock_bh(&policy->lock);

		mipv6_change_sa_index(policy->auth_sa_idx, coa, MIPV6_IPSEC_HA_IN);
		mipv6_change_sa_index(policy->esp_sa_idx, coa, MIPV6_IPSEC_HA_IN);
		mipv6_change_sa_index(policy->comp_sa_idx, coa, MIPV6_IPSEC_HA_IN);

		/* clean up */
		write_unlock_bh(&policy->lock);
		ipsec_sp_put(policy);
	} else {
#ifdef CONFIG_IPV6_MOBILITY_DEBUG
		printk(KERN_INFO "%s(%d): policy is NULL.\n", __FUNCTION__, __LINE__);
#endif
	}
	return 1;
}

int mipv6_mn_change_tunnel_ipsec(struct mipv6_bul_entry *entry)
{
	/*
	 * Phase 1: Change the SPD for the return routability
	 * o SAD3: SRC=CoA DST=HA
	 * o SPD3: SRC=HoA DST=any tunnel mode SGW=HA  proto = MH
	 * o SPD4: SRC=any DST=HoA tunnel mode SGW=CoA proto = MH
	 * o SAD4: SRC=HA  DST=CoA
	 */
	__mipv6_mn_change_tunnel_ipsec_by_proto(entry, IPPROTO_MOBILITY);

	/*
	 * Phage 2: Change the SPD for the Payload Packet
	 * o SPD7: SRC=HoA DST=any tunnel mode SGW=HA  proto = any
	 * o SAD7: SRC=CoA DST=HA
	 * o SPD8: SRC=any DST=HoA tunnel mode SGW=CoA proto = any
	 * o SAD8: SRC=HA  DST=CoA
	 */
	__mipv6_mn_change_tunnel_ipsec_by_proto(entry, 0);

	/*
	 * XXX: IPsec for Home Registration and Mobile Prefix Solicitation
	 * now unchanged.
	 */
	return 1;
}

int __mipv6_ha_change_tunnel_ipsec_by_proto(struct mipv6_bcache_entry *entry, __u8 proto)
{
	struct ipsec_sp *policy = NULL;

	struct in6_addr dst;
	struct in6_addr src;
	struct in6_addr *coa = &entry->coa;

	/*
	 * Phase 1: Change the following SA/SPD
	 *
	 * o SPD4: SRC=any DST=HoA tunnel mode SGW=CoA proto = MH
	 * o SPD8: SRC=any DST=HoA tunnel mode SGW=CoA proto = any
	 * o SAD4: SRC=HA  DST=CoA
	 * o SAD8: SRC=HA  DST=CoA
	 */ 
	memcpy(&dst, &entry->home_addr, sizeof(struct in6_addr));
	memset(&src, 0, sizeof(struct in6_addr));

	policy = mipv6_find_spd(&dst, &src, 128, 0, proto, IPSEC_MODE_TUNNEL);
	if (policy) {
		/*
		 * change the policy:
		 * 1) change the SGW to the current coa;
		 * 2) change the SADB destination address;
		 */

		write_lock_bh(&policy->lock);

		mipv6_change_sa_index(policy->auth_sa_idx, coa, MIPV6_IPSEC_HA_OUT);
		mipv6_change_sa_index(policy->esp_sa_idx, coa, MIPV6_IPSEC_HA_OUT);
		mipv6_change_sa_index(policy->comp_sa_idx, coa, MIPV6_IPSEC_HA_OUT);

		/* clean up */
		write_unlock_bh(&policy->lock);
		ipsec_sp_put(policy);
	}
	/*
	 * Phase 2: Change the following SA/SPD
	 *
	 * o SPD3: SRC=HoA DST=any tunnel mode SGW=HA  proto = MH
	 * o SPD7: SRC=HoA DST=any tunnel mode SGW=HA  proto = any
	 * o SAD3: SRC=CoA DST=HA
	 * o SAD7: SRC=CoA DST=HA
	 */
	policy = NULL;
	memset(&dst, 0, sizeof(struct in6_addr));
	memcpy(&src, &entry->home_addr, sizeof(struct in6_addr));

	policy = mipv6_find_spd(&dst, &src, 0, 128, proto, IPSEC_MODE_TUNNEL);
	if (policy) {
		/*
		 * change the policy:
		 * 1) change the SGW from the current coa;
		 * 2) change the SADB destination address;
		 */

		write_lock_bh(&policy->lock);

		mipv6_change_sa_index(policy->auth_sa_idx, coa, MIPV6_IPSEC_HA_IN);
		mipv6_change_sa_index(policy->esp_sa_idx, coa, MIPV6_IPSEC_HA_IN);
		mipv6_change_sa_index(policy->comp_sa_idx, coa, MIPV6_IPSEC_HA_IN);

		/* clean up */
		write_unlock_bh(&policy->lock);
		ipsec_sp_put(policy);
	}
	return 1;
}

int mipv6_ha_change_tunnel_ipsec(struct mipv6_bcache_entry *entry)
{
	/*
	 * Phase 1: Change the SPD for the return routability
	 * o SAD3: SRC=CoA DST=HA
	 * o SPD3: SRC=HoA DST=any tunnel mode SGW=HA  proto = MH
	 * o SPD4: SRC=any DST=HoA tunnel mode SGW=CoA proto = MH
	 * o SAD4: SRC=HA  DST=CoA
	 */
	__mipv6_ha_change_tunnel_ipsec_by_proto(entry, IPPROTO_MOBILITY);

	/*
	 * Phage 2: Change the SPD for the Payload Packet
	 * o SPD7: SRC=HoA DST=any tunnel mode SGW=HA  proto = any
	 * o SAD7: SRC=CoA DST=HA
	 * o SPD8: SRC=any DST=HoA tunnel mode SGW=CoA proto = any
	 * o SAD8: SRC=HA  DST=CoA
	 */
	__mipv6_ha_change_tunnel_ipsec_by_proto(entry, 0);

	/*
	 * XXX: IPsec for Home Registration and Mobile Prefix Solicitation
	 * now unchanged.
	 */
	return 1;
}

/*
 * check the Security Policy for Home Registration
 */
int mipv6_home_regist_with_esp(struct in6_addr *ha, struct in6_addr *hoa)
{

	struct ipsec_sp *policy;
	int ret = 0;

#ifdef CONFIG_IPV6_MOBILITY_DEBUG
	__u8 buf[INET6_ADDRSTRLEN];
	printk(KERN_INFO "%s(%d): ha = %s\n", __FUNCTION__, __LINE__, in6_ntop(ha, buf));
	printk(KERN_INFO "%s(%d): hoa = %s\n", __FUNCTION__, __LINE__, in6_ntop(hoa, buf));
#endif

	policy = mipv6_find_spd(ha, hoa, 128, 128, IPPROTO_MOBILITY, IPSEC_MODE_TRANSPORT);

	if (policy){
		read_lock_bh(policy->lock);

		if (policy->esp_sa_idx)
			ret = 1;

		read_unlock_bh(policy->lock);
		ipsec_sp_put(policy);
	}
#ifdef CONFIG_IPV6_MOBILITY_DEBUG
	if (ret == 1)
		printk(KERN_INFO "%s(%d): Alternate Care-of Address should be added.\n", __FUNCTION__, __LINE__);
#endif
	return ret;
}
